#include <tchar.h>
#include <windows.h>
#include <stdio.h>
#include "../ArchiveManager/ArchiveManager.h"
#include <string>
#include <shlwapi.h>

#ifdef _DEBUG
#pragma comment(lib, "../Debug/ArchiveManager.lib")
#else
#pragma comment(lib, "../Release/ArchiveManager.lib")
#endif

#pragma comment(lib, "shlwapi.lib")

static const WCHAR *archiveNames[] = {
	L"Unspported",
	L"zip",
	L"lzh",
	L"cab", 
	L"rar",
	L"iso",
};

void usage()
{
	puts("CompareArchive [file] [directory] ");
	puts("  [file]      A[JCut@C");
	puts("  [directory] A[JCuWJꂽfBNg");
}

LONGLONG GetFileSizeFromPath(const WCHAR *path)
{
	HANDLE hFile = CreateFile( path, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_READONLY, NULL );
	if ( hFile == INVALID_HANDLE_VALUE ) {
		_tprintf( L"%s is not found", path );
		return 0;
	}

	LARGE_INTEGER size;
	GetFileSizeEx( hFile, (PLARGE_INTEGER)&size);
	CloseHandle( hFile );
	return size.QuadPart;
}


int sha1(const WCHAR *file, BYTE *hash);
bool Compare(const WCHAR *orig, const WCHAR *file)
{
	BYTE hash1[20], hash2[20];
	int res;

	res = sha1( orig, hash1 );
	if (res == -1) {
		puts("error\n");
		return true;
	}
	res = sha1( file, hash2 );
	if (res == -1) {
		puts("error\n");
		return true;
	}

	_tprintf( L"Base: " );
	for (int i = 0; i < 16; i++) { printf("%X", hash1[i]); } printf("\n");
	_tprintf( L"Temp: " );
	for (int i = 0; i < 16; i++) { printf("%X", hash2[i]); } printf("\n");

	printf("  ");
	if ( memcmp(hash1, hash2, 20) ) {
		puts("bad\n");
		return false;
	} else {
		printf("ok\n");
		return true;
	}
}

int _tmain(int argc, _TCHAR* argv[])
{
	if ( argc != 3 ) {
		usage();
		return -1;
	}

	puts("CompareArcive ver 0.1");

	// Check ArchiveManager's version
	DWORD dwVersion;
	if ((dwVersion = GetAMVersion()) != ARCHIVE_MANAGER_VERSION) {
		printf("ArchiveManager.dll version is different\n");
		return -1;
	} else {
		dwVersion = ARCHIVE_MANAGER_VERSION;
	}

	// Print ArchiveManager.dll version
	printf("ArchiveManager.dll ver %d.%d\n", HIWORD(dwVersion), LOWORD(dwVersion));

	// Open archive
	HARCHIVE hArchive = OpenArchive(argv[1]);
	if ( hArchive == NULL ) {
		_tprintf( L"Failed: OpenArchive %s\n", argv[1] );
		return -1;
	}

	// Print archive type
	_tprintf(L"Opened: \"%s\" as [%s]\n", argv[1], archiveNames[GetArchiveType(hArchive)]);

	// Print number of files
	UInt32 num = GetFileCount(hArchive);
	_tprintf(L"Number of files: %d\n\n", num);

	// ArchiveFileName
	TCHAR f[MAX_PATH];
	lstrcpyn( f, argv[1], MAX_PATH );
	TCHAR *file = PathFindFileName( f );

	// Directory(if not '\\' then add '\\')
	std::wstring dir = argv[2];
	if ( dir[dir.size()-1] != L'\\' ) {
		dir += L'\\';
	}
	dir += file;
	dir +=L'\\';

	// Flag
	bool badFile = false;

	const int STR_SIZE = 1024 * 32;
	WCHAR name[STR_SIZE];
	for ( UInt32 i = 0; i < num; i++ ) {
		// Get FileName
		UInt64 size = GetFileSizeFromArchive( hArchive, i );
		DWORD res = GetFilePathFromArchive( hArchive, i, name, STR_SIZE );
		if ( STR_SIZE < res ) {
			printf( "index: %u StrSize: %u i[ꂽt@C܂\n", i, res );
			continue;
		}
	
		// FileName
		_tprintf( L"* %s\n", name );

		// Get Temporary File Name
		WCHAR out[MAX_PATH];
		res = GetTempFileName( dir.c_str(), L"sevenziptest_", 0, out );
		if (!res) {
			_tprintf(L"Error: GetTempFileName\n");
			exit(1);
		}

		// Extract
		//std::wstring outputPath = 
		BOOL bRes = ExtractToFile( hArchive, i, out, NULL, NULL );
		if ( !bRes ) {
			DWORD err = GetAMLastError();
			_tprintf( L"Error: ", name );
			switch ( err ) {
				case ERROR_INVALID_DATA:
					_tprintf( L"A[JCũf[^₵" );
					break;
				case ERROR_CRC:
					_tprintf( L"Bad CRC value" );
					break;
				case ERROR_UNSUPPORTED_TYPE:
					_tprintf( L"T|[gȂA[JCu" );
					break;
				case ERROR_OUTOFMEMORY:
					_tprintf( L"܂" );
					break;
				default:
					_tprintf( L"̑̃G[" );
					break;
			}
			continue;
		}
		
		// Compare
		std::wstring path( dir );
		path += name;
		_tprintf( L"Base: %s\n", path.c_str());
		if ( !Compare( path.c_str(), out ) ) {
			badFile = true;
		}

		// Del
		DeleteFile( out );
	}

	CloseArchive( hArchive );

	return badFile ? 1 : 0;
}
