#include <windows.h>
#include <stdio.h>
#include <process.h>
#include <stdarg.h>
#include <ctype.h>

#include <imagehlp.h>

#include "ffrutils.h"

#pragma comment(lib, "dbghelp.lib")
#pragma comment(lib, "imagehlp.lib")

#define	FFRUTILS_DEBUG_REG_DUMP_LEN	16

#define d2void(v)	*(void**)(&(v))

static char *DumpMemory(void* pMemory, char *memline, WORD wBytes);

PIMAGEHLP_SYMBOL pSym;

int exception_init()
{
	pSym = (PIMAGEHLP_SYMBOL)GlobalAlloc(GMEM_FIXED, 10000);
	pSym->SizeOfStruct = 10000;
	pSym->MaxNameLength = 10000 - sizeof(IMAGEHLP_SYMBOL);

	return 0;
}

// NEVER CALLED
int exception_destroy()
{
	SymCleanup(GetCurrentProcess());
	GlobalFree(pSym);

	return 0;
}

int GenerateMiniDump(LPEXCEPTION_POINTERS pep)
{
	static LONG counter = 0;
	HANDLE handle;
	MINIDUMP_EXCEPTION_INFORMATION mei;
	SYSTEMTIME		st;
	wchar_t filename[256];

	if (!(ffrutilsopt.debug & FFRUTILS_DEBUG_MINIDUMP)) return 0;

	//
	// setup MINIDUMP_EXCEPTION_INFORMATION
	//
	mei.ClientPointers = FALSE;
	mei.ExceptionPointers = pep;
	mei.ThreadId = GetCurrentThreadId();

	//
	// filename
	//
	GetLocalTime(&st);
	_snwprintf_s(filename,  _countof(filename), _TRUNCATE, L"%s\\minidump_%04d%02d%02d%02d%02d%02d_%x.dmp",
		ffrutilsopt.debug_dir,
		st.wYear, st.wMonth, st.wDay,
		st.wHour, st.wMinute, st.wSecond,
		counter
		);

	InterlockedIncrement(&counter);

	//
	// create a minidump
	//
	if (NULL == (handle = CreateFileW(filename, GENERIC_WRITE,
		0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL)))
	{
		printf("CreateFile(%s): %x\n", filename, GetLastError());
		return 1;
	}

	MiniDumpWriteDump(GetCurrentProcess(), GetCurrentProcessId(), handle, MiniDumpNormal, &mei, NULL, NULL);

	CloseHandle(handle);

	return 0;
}

DWORD Except_Handler(DWORD dwExceptionCode,LPEXCEPTION_POINTERS pep,char *szFunc)
{

	BOOL	EBP_OK = TRUE ;
	DWORD	EBP=0, EIP=0;

	char	szTemp[1024];

	char	*p = NULL;
	MEMORY_BASIC_INFORMATION info;
	char	mod_name[MAX_PATH];

	char	msgtmp[4096];

	DWORD	random=0;

	//xxxxxxxxxxxxxx show debug info xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE,
		"------------------------------------------------------\n"\
		"Function  : %s\n"\
		"Thread ID : 0x%X\n"\
		"Code      : 0x%X\n"\
		"Address   : 0x%p\n"\
		"------------------------------------------------------\n",
			szFunc,GetCurrentThreadId(),dwExceptionCode,pep->ExceptionRecord->ExceptionAddress);

	dbgoutput("%s", msgtmp);

	//xxxxxxxxxxxxxx Registers xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "\nRegisters:\n\n");

	//set eax
	DumpMemory(d2void(pep->ContextRecord->Eax), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sEAX 0x%08lX : %s\n", msgtmp, pep->ContextRecord->Eax, szTemp);

	//set ebx
	DumpMemory(d2void(pep->ContextRecord->Ebx), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sEBX 0x%08lX : %s\n", msgtmp,pep->ContextRecord->Ebx, szTemp);

	//set ecx
	DumpMemory(d2void(pep->ContextRecord->Ecx), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sECX 0x%08lX : %s\n", msgtmp,pep->ContextRecord->Ecx, szTemp);

	//set edx
	DumpMemory(d2void(pep->ContextRecord->Edx), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sEDX 0x%08lX : %s\n", msgtmp,pep->ContextRecord->Edx, szTemp);

	//set edi
	DumpMemory(d2void(pep->ContextRecord->Edi), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sEDI 0x%08lX : %s\n", msgtmp,pep->ContextRecord->Edi, szTemp);

	//set esi
	DumpMemory(d2void(pep->ContextRecord->Esi), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sESI 0x%08lX : %s\n", msgtmp,pep->ContextRecord->Esi, szTemp);

	//set eip
	DumpMemory(d2void(pep->ContextRecord->Eip), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sEIP 0x%08lX : %s\n", msgtmp,pep->ContextRecord->Eip, szTemp);

	//set ebp
	DumpMemory(d2void(pep->ContextRecord->Ebp), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sEBP 0x%08lX : %s\n", msgtmp,pep->ContextRecord->Ebp, szTemp);

	//set esp
	DumpMemory(d2void(pep->ContextRecord->Esp), szTemp, FFRUTILS_DEBUG_REG_DUMP_LEN);
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%sESP 0x%08lX : %s\n", msgtmp,pep->ContextRecord->Esp, szTemp);

	dbgoutput("%s", msgtmp);

	//xxxxxxxxxxxxxx Call Stack xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
#if 0
	_snprintf_s(szCallStack, sizeof(szCallStack), _TRUNCATE, "CallStack:\n\n");

	EIP = pep->ContextRecord->Eip;
	EBP = pep->ContextRecord->Ebp;

	do{
		if( (DWORD)EBP & 3 )
			EBP_OK = FALSE ;
		
		if( EBP_OK && IsBadReadPtr( (void*)EBP, 8 ) )
			EBP_OK = FALSE ;
		
		if( EBP_OK ){

			dwCaller = (DWORD *)(EIP ? (BYTE *)EIP : *((BYTE **)EBP + 1));
			EBP = EIP ? EBP : *(DWORD*)EBP ;
			
			if( EIP )
				EIP = 0 ; // So it is considered just once

			memset(szCaller,0,sizeof(szCaller));
			_snprintf_s(szCaller, sizeof(szCaller), _TRUNCATE, "0x%08p\n", dwCaller);
			_snprintf_s(szCallStack,sizeof(szCaller), _TRUNCATE, "%s%s",szCallStack,szCaller);

		}else
			break ;
	}
	while( TRUE ) ;
#endif 

	//xxxxxxxxxxxxxx Call Stack xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "\nCallStack:\n\n");
	if (1)
	{
		HANDLE h;
		static int syminit = 0;
		STACKFRAME	stackframe;
		DWORD d;

		memset(&stackframe, 0, sizeof(stackframe));

		stackframe.AddrPC.Mode		= AddrModeFlat;
		stackframe.AddrStack.Mode	= AddrModeFlat;
		stackframe.AddrFrame.Mode	= AddrModeFlat;
		stackframe.AddrPC.Offset	= pep->ContextRecord->Eip;
		stackframe.AddrStack.Offset = pep->ContextRecord->Esp;
		stackframe.AddrFrame.Offset = pep->ContextRecord->Ebp;

		if (!syminit)
		{
			syminit++;
			SymInitialize(GetCurrentProcess(), NULL, TRUE);
		}


		h = GetCurrentProcess();

		//
		// Using StackWalk()
		//
		while (StackWalk(IMAGE_FILE_MACHINE_I386,
			h, GetCurrentThread(),
			&stackframe, 
			NULL,
			NULL,
			SymFunctionTableAccess,
			SymGetModuleBase,
			NULL))
		{
			if (SymGetSymFromAddr(h, stackframe.AddrReturn.Offset, &d, pSym))
			{
				_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%s0x%08lX %s+0x%lX\n", msgtmp, stackframe.AddrReturn.Offset, pSym->Name, d);
			}
			else
			{
				_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%s0x%08lX\n", msgtmp, stackframe.AddrReturn.Offset);
			}

			if (0 == stackframe.AddrReturn.Offset) break;
		}

	}

	dbgoutput("%s", msgtmp);

	//xxxxxxxxxxxxxx Loaded Modules xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
	_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "\nLoaded Modules:\n\n");

	for(p = NULL; VirtualQuery(p, &info, sizeof(info)) == sizeof(info);p += info.RegionSize)
	{
		if ((info.AllocationBase == info.BaseAddress) && 
			(GetModuleFileNameA((HINSTANCE)info.AllocationBase, mod_name, sizeof(mod_name))) ) 
		{
			_snprintf_s(msgtmp, sizeof(msgtmp), _TRUNCATE, "%s%s : 0x%08p\n",msgtmp, mod_name, info.AllocationBase);
		}
	}


	GenerateMiniDump(pep);


	dbgoutput("%s", msgtmp);
	
	return 1;
}

static char *DumpMemory(void* pMemory, char *memline, WORD wBytes)
{
	DWORD	memsize;
	DWORD	cdumppos;
	unsigned char  *cptr;
	unsigned short *sptr;
//	char	*memline=NULL;

	memsize=wBytes*2/*HexDump*/+wBytes/2/*Space*/+wBytes/*Char*/+8/*Etc*/;

//	if ((memline=malloc(memsize+1))==NULL) return NULL;
//
//	memset(memline,0,memsize+1);
	*memline = 0;

	cptr = pMemory;
	sptr = pMemory;

	if (!IsBadReadPtr(pMemory, wBytes))
	{
		int i;

		for (i = 0; i < wBytes / 2; i++)
		{
			_snprintf_s(memline + i * 5, memsize - i * 5, _TRUNCATE, "%04x ", (unsigned int)sptr[i]);
		}

		_snprintf_s(memline + i * 5, memsize - i * 5, _TRUNCATE, ": ");

		cdumppos = (DWORD)strlen(memline);
		memset(memline + cdumppos, 0, 17);

		for (i = 0; i < wBytes ; i++)
		{
			if (isprint((unsigned int)(cptr[i])))
				memline[cdumppos + i] = cptr[i];
			else
				memline[cdumppos + i] = '.';
		}
	} 
	else
	{
		_snprintf_s(memline, memsize, _TRUNCATE, "**** **** **** **** **** **** **** ****  : ****************");
	}

	return memline;
}

