// archive_test.cpp : R\[ AvP[ṼGg |Cg`܂B
//

#include "stdafx.h"
#include "../ArchiveManager/ArchiveManager.h"
#include "../ffrutils/ffrutils.h"

#ifdef _DEBUG
#pragma comment(lib, "../Debug/ArchiveManager.lib")
#else
#pragma comment(lib, "../Release/ArchiveManager.lib")
#endif

#ifdef _DEBUG
#pragma comment(lib, "..\\debug\\ffrutils.lib")
#else
#pragma comment(lib, "..\\release\\ffrutils.lib")
#endif

#define STR_SIZE 256

class CTimer
{
protected:
	static bool FreqInitialized;
	static LARGE_INTEGER nFreq;
	LARGE_INTEGER nBefore;
	LARGE_INTEGER nAfter;
public:
	CTimer()
	{
		if ( !FreqInitialized ) {
			QueryPerformanceFrequency( &nFreq );
			FreqInitialized = true;
		}
		ZeroMemory( &nBefore, sizeof(nBefore) );
		ZeroMemory( &nAfter, sizeof(nAfter) );

		// Get
		QueryPerformanceCounter( &nBefore );
	}

	DWORD GetMili()
	{
		QueryPerformanceCounter( &nAfter );
		return (DWORD)(( nAfter.QuadPart - nBefore.QuadPart) * 1000 / nFreq.QuadPart );
	}
	DWORD GetMicro()
	{
		QueryPerformanceCounter( &nAfter );
		return (DWORD)(( nAfter.QuadPart - nBefore.QuadPart) * 1000000 / nFreq.QuadPart );
	}
};

bool CTimer::FreqInitialized = false;
LARGE_INTEGER CTimer::nFreq;

static const WCHAR *archiveNames[] = {
	L"Unspported",
	L"zip",
	L"lzh",
	L"cab", 
	L"rar",
	L"iso",
};

ArchiveType GetArchiveFromString(const WCHAR *str)
{
	for ( int i = 1; i < sizeof(archiveNames)/sizeof(*archiveNames); i++ ) {
		if ( !lstrcmp(str, archiveNames[i]) ) {
			return i;
		}
	}
	return ARCHIVE_NONE;
}

void usage()
{
	puts("test3 [type] [target archive file]\n"
		 "   test all APIs\n"
		 );
}

typedef struct {
	UInt64 CallbackCount;
} CallbackParam;

BOOL  ExtractCallback(LPVOID param, UInt64 completed, UInt64 total)
{
	CallbackParam *cp = (CallbackParam*)param;
	cp->CallbackCount++;
	return TRUE;
}

void hexdump(const void *p, DWORD size);

void PrintErrorString(const WCHAR *function, DWORD err)
{
	const WCHAR *p;
	switch ( err ) {
		case ERROR_INVALID_DATA:
			p = L"A[JCũf[^sł";
			break;
		case ERROR_CRC:
			p = L"CRC̒lv܂";
			break;
		case ERROR_UNSUPPORTED_TYPE:
			p = L"T|[gȂނ̃A[JCuEA[JCũt@C";
			break;
		case ERROR_OUTOFMEMORY:
			p = L"s邽߂̃܂";
			break;
		case ERROR_INTERNAL_ERROR:
			p = L"G[";
			break;
		default:
			//p = L"̑̃G[";
			_tprintf( L"Error: %s %u\n", function, err );
			return;
	}
	_tprintf( L"Error: %s %s\n", function, p );
}

int _tmain(int argc, _TCHAR* argv[])
{
	_wsetlocale( LC_ALL, L"" );

	ffrutils_init();

	ffrutilsopt.debug = FFRUTILS_DEBUG_MINIDUMP | FFRUTILS_DEBUG_LOG;

	__try {
		HARCHIVE hArchive;
		UInt32 num, i;
		DWORD dwVersion;

		if (argc != 3) {
			usage();
			return 0;
		}

	#if 0	// skip
		// Check ArchiveManager's version
		if ((dwVersion = GetAMVersion()) != ARCHIVE_MANAGER_VERSION) {
			_tprintf(L"ArchiveManager.dll is old version\n");
		} else {
			dwVersion = ARCHIVE_MANAGER_VERSION;
		}
	#endif

		ArchiveType type = GetArchiveFromString( argv[1] );
		if ( type == ARCHIVE_NONE ) {
			_tprintf( L"Unspported archive\n" );
			return 1;
		}

		// Open archive
		CTimer openTimer;
		hArchive = OpenArchiveAs(argv[2], type);
		if (hArchive == NULL) {
			PrintErrorString( L"OpenArchiveAs", GetAMLastError() );
			return 1;
		}
		_tprintf( L"  Opening: %u micro sec\n", openTimer.GetMicro() );

		// Print number of files
		CTimer getCount;
		num = GetFileCount(hArchive);
		_tprintf(L"  Files: %u\n", num);
		_tprintf(L"  GetFileCount: %u mili sec\n", getCount.GetMili() );

		// Follow files of archive
		for (i = 0; i < num; i++) {
			// Get file path
			WCHAR path[STR_SIZE] = { 0 };
			DWORD strPathSize = GetFilePathFromArchive(hArchive, i, path, STR_SIZE);

			// Get file name
			WCHAR name[STR_SIZE] = { 0 };
			DWORD strSize = GetFileNameFromArchive(hArchive, i, name, STR_SIZE);
			UInt64 size = GetFileSizeFromArchive(hArchive, i);

			// Check size
			if (STR_SIZE < strSize) {
				_tprintf(L"Index [%d]: File name is too long\n", i);
			}

			// Extract to Handle
			WCHAR tempName[MAX_PATH];
			GetTempFileName( L".\\", L"test3_", 0, tempName );

			// CallbackParam( count, .. )
			CallbackParam cp;
			cp.CallbackCount = 0;

			HANDLE hFile = CreateFile(tempName, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS , FILE_FLAG_DELETE_ON_CLOSE, NULL);
			if (hFile == INVALID_HANDLE_VALUE) {
				_tprintf(L"Fialed to CreateFile\n");
			} else {
				UInt64 size = GetFileSizeFromArchive( hArchive, i );

				CTimer extractTimer;
				BOOL res = ExtractToHandle( hArchive, i, hFile, ExtractCallback, &cp );
				_tprintf( L"  ExtractToHandel - FileSize:%I64u CallbackCount:%I64u  %u mili sec\n", size, cp.CallbackCount, extractTimer.GetMili() );
				if ( res ) {
					CloseHandle(hFile);
				} else {
					PrintErrorString( L"ExtractToHandle", GetAMLastError() );
				}
			}

		}

		// Close handle
		CloseArchive(hArchive);

		return 0;
	} __except(Except_Handler(GetExceptionCode(),GetExceptionInformation(),__FUNCTION__)) {
		exit(2);
	}
}


void hexdump(const void *p, DWORD size)
{
	for (DWORD i = 0; i < size; i++) {
		for (int j = 0; j < 20 && i < size; i++, j++) {
			_tprintf(L"%2x ", *(((BYTE*)p + i)));
		}
		_tprintf(L"\n");
	}
}
